unit DataObj;

interface
uses Windows,ActiveX,Classes;

type
    TDataRequestNotify=procedure(const DataList:TStringList;IsFirstTime:boolean)of object;
    TFileSource=class(TinterfacedObject,IDropSource)
      function QueryContinueDrag(fEscapePressed:BOOL;grfKeyState:Longint):HResult;stdcall;
      function GiveFeedback(dwEffect:Longint):HResult;stdcall;
    end;
    TDataObject=class(TinterfacedObject,IDataObject)
    private
     FFileList:TStringList;
     FDataRequest:TDataRequestNotify;
     FIsFirstTime:boolean;
     FDropPoint:TPoint;
     FInClient:boolean;
     function StorageSize:integer;
    public
     constructor Create(const AFileList:TStringList;ADataRequest:TDataRequestNotify;DropPoint:TPoint;InClient:boolean);
     destructor Destroy;override;
     {IDataObject}
     function GetData(const formatetcIn:TFormatEtc;out medium:TStgMedium):HResult;stdcall;
     function GetDataHere(const formatetc:TFormatEtc;out medium:TstgMedium):HResult;stdcall;
     function QueryGetData(const formatetc:TFormatEtc):HResult;stdcall;
     function GetCanonicalFormatEtc(const formatetc:TFormatEtc;out formatetcOut:TFormatEtc):HResult;stdcall;
     function SetData(const formatetc:TFormatEtc;var medium:TStgMedium;fRelease:BOOL):HResult;stdcall;
     function EnumFormatEtc(dwDirection:Longint;out enumFormatEtc:IEnumFormatEtc):HResult;stdcall;
     function DAdvise(const formatetc:TFormatEtc;advf:Longint;const advSink:IAdviseSink;out dwConnection:Longint):HResult;stdcall;
     function DUnadvise(dwConnection:Longint):HResult;stdcall;
     function EnumDAdvise(out enumAdvise:IEnumStatData):HResult;stdcall;
    end;

implementation
  uses ShlObj,SysUtils;

const
     DataFormatCount=1;

type
    PFormatList=^TFormatList;
    TFormatList=array [0..DataFormatCount-1] of TFormatEtc;

var
   DataFormats:TFormatList;

type
    TEnumFormatEtc=class(TinterfacedObject,IEnumFormatEtc)
    private
     FFormatList:PFormatList;
     FFormatCount:Integer;
     Findex:Integer;
    public
     constructor Create(FormatList:PFormatList;FormatCount,Index:Integer);
     function Next(celt:Longint;out elt;pceltFetched:PLongint):HResult;stdcall;
     function Skip(celt:Longint):HResult;stdcall;
     function Reset:HResult;stdcall;
     function Clone(out enum:IEnumFormatEtc):HResult;stdcall;
    end;

function TFileSource.QueryContinueDrag(fEscapePressed:BOOL;grfKeyState:Longint):HResult;
begin
     if fEscapePressed then Result:=DRAGDROP_S_CANCEL
     else 
        if (grfKeyState and MK_LBUTTON)=0 then Result:=DRAGDROP_S_DROP
        else Result:=S_OK;
end;

function TFileSource.GiveFeedback(dwEffect:Longint):HResult;
begin
     case dwEffect of
          DROPEFFECT_NONE,
          DROPEFFECT_COPY,
          DROPEFFECT_LINK,
          DROPEFFECT_SCROLL:Result:=DRAGDROP_S_USEDEFAULTCURSORS;
          else Result:=S_OK;
     end;
end;
{---------------------------}
constructor TEnumFormatEtc.Create(FormatList:PFormatList;FormatCount,Index:Integer);
begin
     inherited Create;
     FFormatList:=FormatList;
     FFormatCount:=FormatCount;
     FIndex:=Index;
end;

function TEnumFormatEtc.Next(celt:Longint;out elt;pceltFetched:Plongint):HResult;
var i:integer;
begin
     i:=0;
     while (i<celt)and(FIndex<FFOrmatCount) do
     begin
      TFormatList(elt)[i]:=FFOrmatList[FIndex];
      inc(FIndex);
      inc(i);
     end;
     if pceltFetched<>nil then pceltFetched^:=i;
     if i=celt then Result:=S_OK
     else Result:=S_FALSE;
end;

function TEnumFormatEtc.Skip(celt:Longint):HResult;
begin
     if celt<=FFormatCount-FIndex then 
     begin
      FIndex:=FIndex+celt;
      Result:=S_OK;
     end
     else begin
           FIndex:=FFOrmatCount;
           Result:=S_FALSE;
     end;
end;

function TEnumFormatEtc.Reset:HResult;
begin
     FIndex:=0;
     Result:=S_OK;
end;

function TEnumFormatEtc.Clone(out enum:IEnumFormatEtc):HResult;
begin
     enum:=TEnumFormatEtc.Create(FFormatList,FFormatCount,FIndex);
     Result:=S_OK;
end;
{----------------------------}
constructor TDataObject.Create(const AFileList:TStringList;ADataRequest:TDataRequestNotify;DropPoint:TPoint;InClient:boolean);
begin
     inherited Create;
     FFileList:=TStringList.Create;
     FFileList.Assign(AFileList);
     FDataRequest:=ADataRequest;
     FIsFirstTime:=True;
     FDropPoint:=DropPoint;
     FInClient:=InClient;
end;

destructor TDataObject.Destroy;
begin
     if Assigned(FFileList)
                           then FFileList.Free;
     inherited Destroy;
end;

function TDataObject.StorageSize:integer;
var i:integer;
begin
     Result:=sizeof(TDropFiles);{Double-terminated null}
     if FFileList.Count>0 then 
       for i:=0 to FFIleList.Count-1 do
             Result:=Result+length(FFileList[i])+1;
     inc(Result);
end;

function TDataObject.GetData(const formatetcIn:TFormatEtc;out medium:TStgMedium):HResult;
var Data:HGlobal;
    P:pchar;
    i,n:integer;
    DF:PDROPFIles;
    S:string;
begin
     Result:=DV_E_FORMATETC;
     medium.tymed:=0;
     medium.hGlobal:=0;
     medium.unkForRelease:=nil;
     with formatetcIn do
       if (cfFormat=CF_HDROP) and
          (dwAspect=DVASPECT_CONTENT) and
          (tymed=TYMED_HGLOBAL) then 
       begin
         if Assigned(FDataRequest) then FDataRequest(FFileList,FIsFirstTime);
         FIsFirstTime:=False;
         Data:=GlobalAlloc(GMEM_SHARE or GMEM_MOVEABLE or GMEM_ZEROINIT,StorageSize);
         if (Data<>0) and (FFileList.Count>0) then 
         begin
           DF:=GlobalLock(Data);
           if Assigned(DF) then 
           begin
             DF.pFiles:=sizeof(TDropFiles);
             DF.pt:=FDropPoint;
             DF.fNC:=FInClient;
             DF.fWide:=False;
             P:=pchar(DF);
             n:=sizeof(TDropFiles);
             for i:=0 to FFileList.Count-1 do
             begin
               S:=FFileList[i];
               System.Move(S[1],P[n],length(S));
               n:=n+length(s)+1;
             end;
             GlobalUnlock(Data);
           end;
           medium.tymed:=TYMED_HGLOBAL;
           medium.hGlobal:=Data;
           Result:=S_OK;
         end;
       end;
end;

function TDataObject.GetDataHere(const formatetc:TFormatEtc;out medium:TStgMedium):HResult;
begin
     Result:=DV_E_FORMATETC;
end;

function TDataObject.QueryGetData(const formatetc:TFormatEtc):HResult;
begin
     Result:=DV_E_FORMATETC;
     with formatetc do
      if dwAspect=DVASPECT_CONTENT then 
        if (cfFormat=CF_HDROP) and (tymed=TYMED_HGLOBAL) then Result:=S_OK;
end;

function TDataObject.GetCanonicalFormatEtc(const formatetc:TFormatEtc;out formatetcOut:TFormatEtc):HResult;
begin
     formatetcOut.ptd:=nil;
     Result:=E_NOTIMPL;
end;

function TDataObject.SetData(const formatetc:TFormatEtc;var medium:TStgMedium;fRelease:BOOL):HResult;
begin
     Result:=E_NOTIMPL;
end;

function TDataObject.EnumFormatEtc(dwDirection:Longint;out enumFormatEtc:IEnumFormatEtc):HResult;
begin
     if dwDirection=DATADIR_GET then 
     begin
       enumFormatEtc:=TEnumFormatEtc.Create(@DataFormats,DataFormatCount,0);
       Result:=S_OK;
     end
     else begin
             enumFormatEtc:=nil;
             Result:=E_NOTIMPL;
          end;
end;

function TDataObject.DAdvise(const formatetc:TFormatEtc;advf:Longint;const advSink:IAdviseSink;out dwConnection:Longint):HResult;
begin
     Result:=OLE_E_ADVISENOTSUPPORTED;
end;

function TDataObject.DUnadvise(dwConnection:Longint):HResult;
begin
     Result:=OLE_E_ADVISENOTSUPPORTED;
end;

function TDataObject.EnumDAdvise(out enumAdvise:IEnumStatData):HResult;
begin
     Result:=OLE_E_ADVISENOTSUPPORTED;
end;
{--------------------------}

initialization
  DataFormats[0].cfFormat:=CF_HDROP;
  DataFormats[0].ptd:=nil;
  DataFormats[0].dwAspect:=DVASPECT_CONTENT;
  DataFormats[0].lindex:=-1;
  DataFormats[0].tymed:=TYMED_HGLOBAL;
  OleInitialize(nil);

finalization
  OleUnInitialize;

end.
